%-------------------------------------------------------------------------%
% Copyright (c) 2018, Modenese L. and Erica Montefiori                    %
%                                                                         %
% Licensed under the Apache License, Version 2.0 (the "License");         %
% you may not use this file except in compliance with the License.        %
% You may obtain a copy of the License at                                 %
% http://www.apache.org/licenses/LICENSE-2.0.                             %
%                                                                         % 
% Unless required by applicable law or agreed to in writing, software     %
% distributed under the License is distributed on an "AS IS" BASIS,       %
% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         %
% implied. See the License for the specific language governing            %
% permissions and limitations under the License.                          %
%                                                                         %
%    Author:   Luca Modenese, February 2016                               %
%    email:    l.modenese@sheffield.ac.uk                                 % 
% ----------------------------------------------------------------------- %
% This script is used to calculate the axes of the tibio-talar and subtalar
% joints. The tibio-talar axis is computed by fitting a cylinder to the
% talar troclea's surface, while the subtalar axis is identified by fitting
% two spheres to the talo-calcanear and talo-navicular articular surfaces
% and connecting them with an axis.
% The code performs a least square fitting of the cylinder
% using the Least Squares Geometric Elements library MATLAB package (LSGE), 
% which is not included and needs to be downloaded at 
% http://www.eurometros.org/gen_report.php?category=distributions&pkey=14&subform=yes
% and corresponding functions need to be added to the MATLAB path, for
% instance copying them in the ./Fitting_function directory.
%
% The scripts requires the user to specify a folder where the geometries
% are stored and assumes that it will contain two files named:
% - talocalcaneal.stl
% - talonavicular.stl
% - talartrochlea.stl
% that it will use for the fitting procedure.
% The script has to be run for each side.
% ----------------------------------------------------------------------- %
% Last modified
% October 2016, Erica Montefiori
% Added printing of the output to Fitted_Geometries.txt file compatible with NMS Builder
% landmark clouds import tool. 
% April 2018, Erica Montefiori
% Updated cylinder and sphere plots.
% ----------------------------------------------------------------------- %
% initialize
clear;clc;
close all; fclose all; 

% add functions to path
addpath(genpath('./Fitting_functions'))
addpath(genpath('./ReadWrite_functions'))


% the user selects the directory
geometry_folder = uigetdir('./geometries','Select the folder containing the geometries');


%% set up log file
log_file = fullfile(geometry_folder,'fittingResults.txt');
% create file and clean it
fid = fopen(log_file,'w+');
fclose all;
% start logging
diary(log_file);
diary on

%% get geometry files
TaloCalcn_file = fullfile(geometry_folder, 'talocalcaneal.stl');
TaloNavic_file = fullfile(geometry_folder, 'talonavicular.stl');
Talus_trochlea = fullfile(geometry_folder, 'talartrochlea.stl');

%log of where the data are
display('Processing data from folder: ')
display(char(geometry_folder));
display(' ');

% initialize figure
h = figure('Position', [1 41 1366 652]);

%% ====== fit sphere to talus-calcaneus joint surface (talus side) =====
% read the file
[v, f] = stlread(TaloCalcn_file);
[v, ~]=patchslim(v, f);
[Center_LSE,Radius_LSE] = sphereFit(v);

display('======== TALUS-CALCANEUS ==========')
display(['Sphere centre: ', num2str(Center_LSE)]);
display(['Sphere radius: ', num2str(Radius_LSE)]);

% plot results
subplot(1,3,1)
% plot geometry
plot3(v(:,1),v(:,2),v(:,3),'.r');hold on
title({'TALUS-CALCANEUS';['Centre: ', num2str(Center_LSE)]; ['Radius: ', num2str(Radius_LSE)]}, 'FontWeight', 'bold')
% plot sphere
sphere_res = 49;
[X,Y,Z] = sphere(sphere_res);
s=surf(Center_LSE(1)+X*Radius_LSE,Center_LSE(2)+Y*Radius_LSE,Center_LSE(3)+Z*Radius_LSE,'FaceColor',[0 0 1]);
hold on
axis equal, grid on

% update names
updPMat(:,1) = Center_LSE;
colheaders_upd{1} =  ('Calc_Sphere_Centre');

clear v


%% ====== fit sphere to talus-navicularis joint surface (navicularis side) =====
% read the file
[v, f] = stlread(TaloNavic_file);
[v, ~]=patchslim(v, f);
[Center_LSE,Radius_LSE] = sphereFit(v);

display('======== TALUS-NAVICULAR ==========')
display(['Sphere centre: ', num2str(Center_LSE)]);
display(['Sphere radius: ', num2str(Radius_LSE)]);

% plot sphere
subplot(1,3,2)
% plot geometry
plot3(v(:,1),v(:,2),v(:,3),'.r'); hold on
title({'TALUS-NAVICULAR';['Centre: ', num2str(Center_LSE)]; ['Radius: ', num2str(Radius_LSE)]}, 'FontWeight', 'bold')
% plot sphere
sphere_res = 49;
[X,Y,Z] = sphere(sphere_res);
s=surf(Center_LSE(1)+X*Radius_LSE,Center_LSE(2)+Y*Radius_LSE,Center_LSE(3)+Z*Radius_LSE,'FaceColor',[0 0 1]); 

hold on
axis equal, grid on

% update names
updPMat(:,2) = Center_LSE;
colheaders_upd{2} =  ('Navic_Sphere_Centre');

clear v
%% ====== fit cylinder to talus trochlea =====

% read geometry
[v, f] =  stlread(Talus_trochlea);
[v, f] = patchslim(v, f);


%== FIND FIRST GUESSES ====
% guess the point on axis and the radius based on sphere fit
[x0(1:3,1), r0] = sphereFit(v);
a0 = [1  0  0 ]';

% tolerances to set before calling the cylinder fit (decided on a trial and
% error base)
tolp = 10^-6; tolg = 10^-6;
% fitting cylinder to talar throclea
[x0n, an, rn, d, sigmah, conv, Vx0n, Van, urn, GNlog, ... 
          a, R0, R] = lscylinder(v, x0, a0, r0, tolp, tolg);
      
      if conv == 0
          display('Covergence not reached....trying different starting point.')
          [x0(1:3,1)] = mean(v);
          % fitting cylinder to talar throclea
          [x0n, an, rn, d, sigmah, conv, Vx0n, Van, urn, GNlog, ...
              a, R0, R] = lscylinder(v, x0, a0, r0, tolp, tolg);
      end
          
% displaying results
display('======== TALUS TROCHLEA ==========')
display(['Cylinder centre: ', num2str(x0n')]);
display(['Cylinder radius: ', num2str(rn)]);
display(['Cylinder axis point (in talus): ', num2str(x0n'+50*an')]);

subplot(1,3,3);

% showing the mesh of talar trochlea
plot3(v(:,1),v(:,2),v(:,3),'.r');hold on
title({'TALUS TROCHLEA';...
    ['Cyl Centre: ', num2str(x0n')];...
    ['Cyl Radius: ', num2str(rn)];...
    ['Cyl Point on axis: ', num2str(x0n'+50*an')]},...
    'FontWeight', 'bold')

% plotting the cylinder

X1=x0n'+15*an';
X2=x0n'-15*an';
r=rn;
n=40;
cyl_color='b';
closed=1;
lines=1;
Cylinder(X1,X2,r,n,cyl_color,closed,lines);
hold on
axis equal, grid on
clear v

% update names
updPMat(:,3) =  x0n';
colheaders_upd{3} =  ('O_cylinder');
updPMat(:,4) =  x0n'+50*an';
colheaders_upd{4} =  ('P_cylinder');
updPMat(:,5) =  x0n'+an';
colheaders_upd{5} =  ('X_cylinder');

clear v
%% save figures 
set(h, 'PaperPositionMode','auto')
saveas(h,fullfile(geometry_folder,'fittingResults.png'));
saveas(h,fullfile(geometry_folder,'fittingResults.fig'));
% close all

% inform the user
display(' ');
display('The plots and a log fils have been saved in geometry folder.');
diary off;

%% remove library from path
rmpath(genpath('../MatlabFunctions_fit'))



%% creating outputs and output files
% creating the structure with the snapped points
updLandmarkCloud = struct;
updLandmarkCloud.colheaders = colheaders_upd;
updLandmarkCloud.data = updPMat;


% write the updated landmark cloud in the same place where the original
% clous is located
%[~, cloud_name, est] = fileparts(cloud_file);
%[~, stl_name, est] = fileparts(stl_file);
fitted_landmarks_file_name = ['Fitted_geometries.txt'];
writeNMSBuilderLandmarkCloud(updLandmarkCloud,fullfile(geometry_folder,fitted_landmarks_file_name));
%writeNMSBuilderLandmarkCloud(updLandmarkCloud,fullfile(geometry_folder,fitted_landmarks_file_name));
%display(['NMSBuilder file ',fitted_landmarks_file_name,' written in folder: ', cloud_path])
diary off

      