%-------------------------------------------------------------------------%
% Copyright (c) 2018, Modenese L. and Erica Montefiori                    %
%                                                                         %
% Licensed under the Apache License, Version 2.0 (the "License");         %
% you may not use this file except in compliance with the License.        %
% You may obtain a copy of the License at                                 %
% http://www.apache.org/licenses/LICENSE-2.0.                             %
%                                                                         % 
% Unless required by applicable law or agreed to in writing, software     %
% distributed under the License is distributed on an "AS IS" BASIS,       %
% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         %
% implied. See the License for the specific language governing            %
% permissions and limitations under the License.                          %
%                                                                         %
%    Author:   Luca Modenese, February 2016                               %
%    email:    l.modenese@sheffield.ac.uk                                 % 
% ----------------------------------------------------------------------- %
% This script is used to calculate the knee joint axis from stl geometries of
% the femoral condyles. The code performs a least square fitting of a cylinder
% using the Least Squares Geometric Elements library MATLAB package (LSGE), 
% which is not included and needs to be downloaded at 
% http://www.eurometros.org/gen_report.php?category=distributions&pkey=14&subform=yes
% and corresponding functions need to be added to the MATLAB path, for
% instance copying them in the ./Fitting_function directory.
%
% The scripts requires the user to specify a folder where the geometries
% are stored and assumes that it will contain two files named:
% - epicondyle_l.stl
% - epicondyle_r.stl
% that it will use for the fitting procedure.
% If generating a unilateral model, the user is supposed to comment the 
% code related to the contralateral side or errors will be thrown.
% ----------------------------------------------------------------------- %
% Last modified
% October 2016, Erica Montefiori
% Added printing of the output to Fitted_Geometries_side.txt file compatible with NMS Builder
% landmark clouds import tool. Now code browse Right and Left folders
% automatically.
% April 2018, Erica Montefiori
% Modified cylinder and sphere plots.
% ----------------------------------------------------------------------- %

% initialize
clear;clc;
close all; fclose all; 

% add functions to path
addpath(genpath('./Fitting_functions'))
addpath(genpath('./ReadWrite_functions'))

% the user selects the directory
geometry_folder = uigetdir('./geometries','Select the folder containing the geometries');


%% set up log file
log_file = fullfile(geometry_folder,'fittingResults.txt');
% create file and clean it
fid = fopen(log_file,'w+');
fclose all;
% start logging
diary(log_file);
diary on

%% get geometry files
A = fullfile(geometry_folder, 'epicondyle_l.stl');
B = fullfile(geometry_folder, 'epicondyle_r.stl');

%log of where the data are
display('Processing data from folder: ')
display(char(geometry_folder));
display(' ');

% initialize figure
h = figure('Position', [1 41 1366 652]);


%% ====== fit cylinder to distal femur (left side) =====

% read geometry
[v1, f1] =  stlread(A);
[v1, f1] = patchslim(v1, f1);


%== FIND FIRST GUESSES ====
% guess the point on axis and the radius based on sphere fit
[x0(1:3,1), r0] = sphereFit(v1);
a0 = [1  0  0 ]';


% tolerances to set before calling the cylinder fit (decided on a trial and
% error base)
tolp = 10^-6; tolg = 10^-6;
% fitting cylinder 
[x0n, an, rn, d, sigmah, conv, Vx0n, Van, urn, GNlog, ... 
          a, R0, R] = lscylinder(v1, x0, a0, r0, tolp, tolg);
      
      if conv == 0
          display('Covergence not reached....trying different starting point.')
          [x0(1:3,1)] = mean(v1);
          % fitting cylinder with new starting point
          [x0n, an, rn, d, sigmah, conv, Vx0n, Van, urn, GNlog, ...
              a, R0, R] = lscylinder(v1, x0, a0, r0, tolp, tolg);
      end
          
% displaying results
display('======== EPICONDYLE LEFT ==========')
display(['Cylinder centre: ', num2str(x0n')]);
display(['Cylinder radius: ', num2str(rn)]);
display(['Cylinder axis point (in femur): ', num2str(x0n'+50*an')]);
 

subplot(1,2,1);

% showing the results of fitting procedure
plot3(v1(:,1),v1(:,2),v1(:,3),'.r');hold on
title({'EPICONDYLE LEFT';...
    ['Cyl Centre: ', num2str(x0n')];...
    ['Cyl Radius: ', num2str(rn)];...
    ['Cyl Point on axis: ', num2str(x0n'+50*an')]},...
    'FontWeight', 'bold')

% plotting the cylinder
X1=x0n'+35*an';
X2=x0n'-35*an';
r=rn;
n=40;
cyl_color='b';
closed=1;
lines=1;
Cylinder(X1,X2,r,n,cyl_color,closed,lines);
hold on
axis equal, grid on
view([150,15]);
clear v

%% update names
updPMat(:,1) =  x0n';
colheaders_upd{1} =  ('O_cylinder_l');
updPMat(:,2) =  x0n'+50*an';
colheaders_upd{2} =  ('P_cylinder_l');


%% creating outputs and output files
% creating the structure with the snapped points
updLandmarkCloud = struct;
updLandmarkCloud.colheaders = colheaders_upd;
updLandmarkCloud.data = updPMat;


% write the updated landmark cloud in the same place where the original
% cloud is located
%[~, cloud_name, est] = fileparts(cloud_file);
%[~, stl_name, est] = fileparts(stl_file);
fitted_landmarks_file_name = ['Fitted_geometries_l.txt'];
writeNMSBuilderLandmarkCloud(updLandmarkCloud,fullfile(geometry_folder,fitted_landmarks_file_name));
%display(['NMSBuilder file ',fitted_landmarks_file_name,' written in folder: ', cloud_path])


%% read geometry
[v2, f2] =  stlread(B);
[v2, f2] = patchslim(v2, f2);

%== FIND FIRST GUESSES ====
% guess the point on axis and the radius based on sphere fit
[x0(1:3,1), r0] = sphereFit(v2);
a0 = [1  0  0 ]';


% tolerances to set before calling the cylinder fit (decided on a trial and
% error base)
tolp = 10^-6; tolg = 10^-6;
% fitting cylinder 
[x0n, an, rn, d, sigmah, conv, Vx0n, Van, urn, GNlog, ... 
          a, R0, R] = lscylinder(v2, x0, a0, r0, tolp, tolg);
      
      if conv == 0
          display('Covergence not reached....trying different starting point.')
          [x0(1:3,1)] = mean(v2);
          % fitting cylinder using new starting point
          [x0n, an, rn, d, sigmah, conv, Vx0n, Van, urn, GNlog, ...
              a, R0, R] = lscylinder(v2, x0, a0, r0, tolp, tolg);
      end
          
% displaying results
display('======== EPICONDYLE RIGHT ==========')
display(['Cylinder centre: ', num2str(x0n')]);
display(['Cylinder radius: ', num2str(rn)]);
display(['Cylinder axis point (in femur): ', num2str(x0n'+50*an')]);

subplot(1,2,2);

% showing results of fitting procedure
plot3(v2(:,1),v2(:,2),v2(:,3),'.r');hold on
title({'EPICONDYLE RIGHT';...
    ['Cyl Centre: ', num2str(x0n')];...
    ['Cyl Radius: ', num2str(rn)];...
    ['Cyl Point on axis: ', num2str(x0n'+50*an')]},...
    'FontWeight', 'bold')

% plotting the cylinder
X1=x0n'+35*an';
X2=x0n'-35*an';
r=rn;
n=40;
cyl_color='b';
closed=1;
lines=1;
Cylinder(X1,X2,r,n,cyl_color,closed,lines);
hold on
axis equal, grid on
view([-150,15]);
clear v

%% update names
updPMat(:,1) =  x0n';
colheaders_upd{1} =  ('O_cylinder_r');
updPMat(:,2) =  x0n'+50*an';
colheaders_upd{2} =  ('P_cylinder_r');


%% creating outputs and output files
% creating the structure with the snapped points
updLandmarkCloud = struct;
updLandmarkCloud.colheaders = colheaders_upd;
updLandmarkCloud.data = updPMat;


% write the updated landmark cloud in the same place where the original
% cloud is located
%[~, cloud_name, est] = fileparts(cloud_file);
%[~, stl_name, est] = fileparts(stl_file);
fitted_landmarks_file_name = ['Fitted_geometries_r.txt'];
writeNMSBuilderLandmarkCloud(updLandmarkCloud,fullfile(geometry_folder,fitted_landmarks_file_name));
%display(['NMSBuilder file ',fitted_landmarks_file_name,' written in folder: ', cloud_path])


%% save figures 
set(h, 'PaperPositionMode','auto')
saveas(h,fullfile(geometry_folder,'fittingResults.png'));
saveas(h,fullfile(geometry_folder,'fittingResults.fig'));
% close all

% inform the user
display(' ');
display('The plots and a log fils have been saved in geometry folder.');
diary off;

%% remove library from path
rmpath(genpath('../MatlabFunctions_fit'))
      