%-------------------------------------------------------------------------%
% Copyright (c) 2018 Modenese L.                                          %
%                                                                         %
% Licensed under the Apache License, Version 2.0 (the "License");         %
% you may not use this file except in compliance with the License.        %
% You may obtain a copy of the License at                                 %
% http://www.apache.org/licenses/LICENSE-2.0.                             %
%                                                                         % 
% Unless required by applicable law or agreed to in writing, software     %
% distributed under the License is distributed on an "AS IS" BASIS,       %
% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         %
% implied. See the License for the specific language governing            %
% permissions and limitations under the License.                          %
%                                                                         %
%    Author:   Luca Modenese, May 2016                                    %
%    email:    l.modenese@sheffield.ac.uk                                 % 
% ----------------------------------------------------------------------- %
%
% This function to modify the joints of an OpenSim, which is thought to be
% used after a model is produced by NMSBuilder. Joints and coordinates of 
% such a model will not have the same names as gait2392, for instance. 
% This function allows to modify: 1) the joint name, 2) the rotations order
% 3) the coordinates names. This is possible only on already existing joints. 
% This function does not add degrees of freedom.
%
% IMPORTANT: this function has been tested only with CustomJoints and it
% was designed to be used within the workflow described in:
% Modenese, L., Montefiori, E., Wang, A., Wesarg, S., Viceconti, M., Mazz, C., 2018. 
% Investigation of the dependence of joint contact forces on musculotendon parameters 
% using a codified workflow for image-based modelling. Journal of Biomechanics. 73, 108-118 
%
% Inputs:
%    - osimModel: an openSim model (java object).
%    - childBodyName: a Matlab string identifying the child of the joint of
%                     interest;
%    - jointDesiredName: the desidered name for the joint of interest;
%    - expectedJointType: the mechanical "type" of joint of interest. This
%                       can be 'spherical', 'hinge' or 'free' and HAS to
%                       match the type of joint in the model, otherwise an
%                       error is thrown.
%    - rotationOrder: a string with three characters defining the rotation
%                     order for the joint, e.g. 'zxy'. Three axes have to
%                     be specified also for hinge joints.
%    - coordsNames: names of the coordinates that the users wants to assign
%                   to the rotation axes. These must be provided in
%                   corresponding order to the rotationOrder, i.e. the
%                   first coordinate will correspond to the first axis of
%                   rotation and so on.
%
% Outputs:
%    - upd_osimModel: openSim model with updated joint features.
%--------------------------------------------------------------------------
% last modified: 25/04/2018, Luca Modenese
% Added comments.
%--------------------------------------------------------------------------

function upd_osimModel = updBuilderJoints2Osim(osimModel, childBodyName, jointDesiredName, expectedJointType, rotationOrder, coordsNames)

% importing libraries
import org.opensim.modeling.*

% TO DO: add a check: the function work only with custom joints

%=========== EXTRACTING THE INPUTS =============
% this allows to feed in a string if a single coordinate is specified
if ~iscell(coordsNames)
    coordsNames = {coordsNames};
end

% distinguishing the type of joint
switch expectedJointType
    case 'free'
        rot_trans = [1 1];
        if size(coordsNames,2)<6; error('This function can only deal with 6dof free joints for now.'); end
    case 'hinge'
        rot_trans = [1 0];
    case 'spherical'
        rot_trans = [1 0];
    otherwise
        error('Supported expected joint types are ''free'', ''hinge'' and ''spherical''.')
end

% extracting the vectors associated with the order of rotation
v(1,1:3) = getAxisVecFromStringLabel(rotationOrder(1));
v(2,1:3) = getAxisVecFromStringLabel(rotationOrder(2));
v(3,1:3) = getAxisVecFromStringLabel(rotationOrder(3));

% TODO: add check on unique axes

% default for translations
v(4:6,1:3) = eye(3);
% =============================================================

% get the current body
if osimModel.getBodySet.getIndex(childBodyName)<0
    upd_osimModel = osimModel;
    return
else
    curr_body = osimModel.getBodySet.get(childBodyName);
end

% get body name
curr_body_name = char(curr_body.getName);

% get Osim joint name associated with that body
% curr_joint_name = getOSJointNameFromChildBodyName(curr_body_name);

% get joint
bodyJoint = curr_body.getJoint();

% change name to OS names
bodyJoint.setName(jointDesiredName);

N_str = size(coordsNames,2);

% update the coordinates!
for n_coord = 0:N_str-1
    
    % the given list is used to extract the desired name
    curr_coord_name = coordsNames{n_coord+1};
    
    %============ COORDINATE SET =======================
    % get coordinate
    curr_coord = bodyJoint.getCoordinateSet.get(n_coord);
    % change coordinate name
    curr_coord.setName(curr_coord_name);
    
    %============SPATIAL TRANSFORM =========================
    % downcast custom joint
    customJ = CustomJoint.safeDownCast(bodyJoint);
    
    % get spatial transform
    jointSpatialTransf = customJ.getSpatialTransform();
    
    % update variable names for as many dof as indicated
    lin_fun = LinearFunction(1, 0);
    const_fun = Constant(0);
    
    %================= TRANSLATIONS ===================
    if rot_trans(1)==1
        % looping through rotations axes
        for n = 1:3
            % get transform axis
            TransAxis_rot = jointSpatialTransf.getTransformAxis(n-1);
            
            % applying standard zxy rotations
            TransAxis_rot = updTransformAxis(TransAxis_rot, v(n,:));
            
            % this will update the coordinate names and assign linear
            % function to those axes with a coordinate associated with,
            % a constant null function to teh others.
            if n<=N_str
                TransAxis_rot = updTransformAxisCoordName(TransAxis_rot, coordsNames{n});
                TransAxis_rot.set_function(lin_fun);
            else
                % no dof on this axes
                TransAxis_rot = updTransformAxisCoordName(TransAxis_rot, '');
                TransAxis_rot.set_function(const_fun);
            end
        end
    end
    
    %================= TRANSLATIONS ====================
    % this section is limited for now.
    % It can only deal with free joints with 6 dof, of which the last 6
    % are translations.
    if rot_trans(2)==1

        % looping through translation axes
        for n = 4:6
            % get transform axis
            TransAxis_trans = jointSpatialTransf.getTransformAxis(n-1);
            % applying standard xyz translations
            TransAxis_trans = updTransformAxis(TransAxis_trans, v(n,:));
            % changing names
            TransAxis_trans = updTransformAxisCoordName(TransAxis_trans, coordsNames{n});
        end
    end
end

% pass model out
upd_osimModel = osimModel;

end