%-------------------------------------------------------------------------%
% Copyright (c) 2018, Modenese L. and Erica Montefiori                    %
%                                                                         %
% Licensed under the Apache License, Version 2.0 (the "License");         %
% you may not use this file except in compliance with the License.        %
% You may obtain a copy of the License at                                 %
% http://www.apache.org/licenses/LICENSE-2.0.                             %
%                                                                         % 
% Unless required by applicable law or agreed to in writing, software     %
% distributed under the License is distributed on an "AS IS" BASIS,       %
% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         %
% implied. See the License for the specific language governing            %
% permissions and limitations under the License.                          %
%                                                                         %
%    Author:   Luca Modenese, May 2016                                    %
%    email:    l.modenese@sheffield.ac.uk                                 % 
% ----------------------------------------------------------------------- %
%
% When the user runs this function, a dialogue allows to select an stl file 
% and a landmark cloud generated by NMSBuilder after that.
% This function finds, for each landmark of the landmark clous, the closest 
% vertex of the stl surface. The function prints a file including the 
% coordinates of the snapped landmarks. The txt file can be read by NMSBuilder 
% and reimported for further analyses.
% The main use of this script is to adjust muscle attachments after affine
% registration. For details, see:
% Modenese, L., Montefiori, E., Wang, A., Wesarg, S., Viceconti, M., Mazz, C., 2018. 
% Investigation of the dependence of joint contact forces on musculotendon parameters 
% using a codified workflow for image-based modelling. Journal of Biomechanics. 73, 108-118 
%
% Inputs:
%    - stl_file: an stl file where the landmark clouds will be snapped.
%    - cloud_file: file containing a NMSBuilder landmarks cloud to snap on a
%                  stl geometry.
%
% Outputs:
%    - updLandmarkCloud file: file containing the updated coordinates of
%                             landmarks snapped to the stl surface
%                             vertices.
%
% ----------------------------------------------------------------------- %
% Last updated: 
% 11/05/2016, LM
% Added readNMSBuilderLandmarkCloud functionality
% bony landmarks have been excluded from the snapped atlas and snapped file's name has been changed.
%
% Last updated: 07/06/2016, Erica Montefiori
% Removed MD-Paedigree folder dependencies.
% ----------------------------------------------------------------------- %

function snapLandmarks

close all
fclose all;clear;clc;
addpath(genpath('./Fitting_functions'))
addpath(genpath('./Snapping_functions'))
addpath(genpath('./ReadWrite_functions'))
%% ================== SETTINGS ============================================
% user specifies the mesh and the 
[stl_file, stl_path] = uigetfile('*.stl','Select the mesh where you want to snap the landmark cloud');
[cloud_file, cloud_path] = uigetfile('*.txt','Select the landmark cloud you want to snap to the surface');
%==========================================================================

%% Acquiring data
% import geometry
[v, f, n, c, stltitle] = stlread(fullfile(stl_path,stl_file));
[v, f]=patchslim(v, f);

% read NMSBuilder lardmark cloud file
landmarks = readNMSBuilderLandmarkCloud(fullfile(cloud_path,cloud_file));
N = size(landmarks.data,2);

% colheaders are the same as the original cloud
colheaders = landmarks.colheaders;

% dimensional check and initialization
if size(colheaders,2)~=N
    error('Column headers and number of landmarks are not consistent.')
else
    % initialization of point matrix
    updPMat = zeros(3,N);
end

%% set up log file
log_file = fullfile(cloud_path,'snap2surf_log.txt');
% create file and clean it
fid = fopen(log_file,'w+');
fclose all;
% start logging
diary(log_file);
diary on

%% snapping points
n_p = 1;

for n = 1:N
    % reading the points
    curr_P_coords = landmarks.data(:,n)';
    curr_P_name = colheaders{n};
    
    % informing user
    disp(['Snapping point (',num2str(n),'/',num2str(N),'): ',curr_P_name]);
    
    if strncmp(curr_P_name,'R',1) || strncmp(curr_P_name,'L',1) || strncmp(curr_P_name,'I',1)
        % Palpated_side point are not to be included in the Muscles_side atlas
        display('Point is recognized as a bone landmark. Skipping.')
        display('----------------------------------------------')
        continue
    end
    
    % check to distinguish between muscle attachments and viapoints
    if ~isempty(strfind(string(curr_P_name),'via')) || strcmp(curr_P_name,'psoas_r orig')  || strcmp(curr_P_name,'psoas_l orig') 
%         display('Point is recognized as viapoint. Not snapping.')
        display('Point is recognized as viapoint or point not to be snapped. Not snapping.')
        display('----------------------------------------------')
        % we do not want to snap on surface the viapoints
        upd_P_coords = curr_P_coords;
    else
        % but we do want to snap all the other points
        
        %========== CODE TAKEN FROM SCRIPTS INTENDED FOR AUTOMATIC DETECTION OF
        %BONY LANDMARKS =======================================================
        
        [upd_P_coords, min_dist] = getClosestPoint(curr_P_coords, v);
        
%         % Material included in getClosestPoint.m
%         % building vector for vectorial difference
%         P_vec = ones(size(v,1),1)*curr_P_coords;
%         % calculate distance between P and all points in mesh
%         dist = (sum((v-P_vec).^2.0, 2)).^0.5;
%         [min_dist, minval]  = min(dist);
%         upd_P_coords = v(minval,:);
        %======================================================================
        % display new coordinates of points
        display(['Moved from [',num2str(curr_P_coords(1)),' ',num2str(curr_P_coords(2)),' ',num2str(curr_P_coords(3)),'] to [',...
            num2str(upd_P_coords(1)),' ',num2str(upd_P_coords(2)),' ',num2str(upd_P_coords(3)),']']);
        display(['Distance from original position: ',num2str(min_dist),' mm']);
        display('----------------------------------------------')
    end
    
    % updating landmark cloud 
    updPMat(:,n_p) = upd_P_coords;
    colheaders_upd{n_p} =  curr_P_name;
    n_p = n_p+1;

end

%% creating outputs and output files
% creating the structure with the snapped points
updLandmarkCloud = struct;
updLandmarkCloud.colheaders = colheaders_upd;
updLandmarkCloud.data = updPMat;

% write the updated landmark cloud in the same place where the original
% clous is located
[~, cloud_name, est] = fileparts(cloud_file);
[~, stl_name, est] = fileparts(stl_file);
snapped_landmarks_file_name = ['Muscles_',stl_name ,'_snapped.txt'];
writeNMSBuilderLandmarkCloud(updLandmarkCloud,fullfile(cloud_path,snapped_landmarks_file_name));
display(['NMSBuilder file ',snapped_landmarks_file_name,' written in folder: ', cloud_path])
diary off

rmpath(genpath('./Fitting_functions'))
rmpath(genpath('./Snapping_functions'))
rmpath(genpath('./ReadWrite_functions'))
end

%-------------------------------------------------------------------------%
% Copyright (c) 2016 Modenese L.                                          %
%                                                                         %
% Licensed under the Apache License, Version 2.0 (the "License");         %
% you may not use this file except in compliance with the License.        %
% You may obtain a copy of the License at                                 %
% http://www.apache.org/licenses/LICENSE-2.0.                             %
%                                                                         % 
% Unless required by applicable law or agreed to in writing, software     %
% distributed under the License is distributed on an "AS IS" BASIS,       %
% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         %
% implied. See the License for the specific language governing            %
% permissions and limitations under the License.                          %
%                                                                         %
%    Author:   Luca Modenese, May 2016                                    %
%    email:    l.modenese@sheffield.ac.uk                                 % 
% ----------------------------------------------------------------------- %
% This function takes as input a point [x y z] and a landmark cloud [N x 3]
% and returns the point from the cloud that is the closest to given point.
% It also returns the distance between the two points (minimum distance).
% Last updated: 11/05/2016
function [closestPoint, min_dist] = getClosestPoint(point, pointCloud)

    function v = checkDimensions(v)
        % checks dimensionality
        if size(v,2)~=3
            if size(v,1)==3
                warndlg('dimensional check before searching for closest point on landmark cloud cannot be performed because only 3 points are present in the surface.');
            else
                v = v';
            end
        end
    end

% check dimensions and arrangement of vectors
point = checkDimensions(point);
pointCloud = checkDimensions(pointCloud);

% create a [N_point x 3]matrix with the coordinates of the point as rows
point_mat = ones(size(pointCloud,1),1)*point;

% calculate distance from all points of the cloud
dist = (sum((pointCloud-point_mat).^2.0, 2)).^0.5;

% get minimum distance
[min_dist, min_dist_ind]  = min(dist);

% extract minimum distance point
closestPoint = pointCloud(min_dist_ind,:);
end
