%-------------------------------------------------------------------------%
% Copyright (c) 2018 Modenese L.                                          %
%                                                                         %
% Licensed under the Apache License, Version 2.0 (the "License");         %
% you may not use this file except in compliance with the License.        %
% You may obtain a copy of the License at                                 %
% http://www.apache.org/licenses/LICENSE-2.0.                             %
%                                                                         % 
% Unless required by applicable law or agreed to in writing, software     %
% distributed under the License is distributed on an "AS IS" BASIS,       %
% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         %
% implied. See the License for the specific language governing            %
% permissions and limitations under the License.                          %
%                                                                         %
%    Author:   Luca Modenese,  2016                                       %
%    email:    l.modenese@sheffield.ac.uk                                 % 
% ----------------------------------------------------------------------- %
% This function takes as input a lower limb model produced by NMSBuilder 
% and updates the names of joints and coordinates and the rotation orders
% in order to match those of gait2392 model.
%
% Inputs:
%    - input OpenSim model: chosen through a dialog menu, it is assumed to
%      be a bilateral model with joints defined as follows:
%      hip: spherical joint; 
%      knee/ankle/subtalar/mtp joint: hinge joint.
%
% Outputs:
%    - an output model is saved in the same folder of the input model, with
%      name [input_model,'_upd_names.osim'].
%
% LIMITATIONS:
% 1) Please not that the joints to be modified have to correspond to the
% definition emplyed in updBuilderJoints2Osim.m. This means that if a
% joint is defined as 'spherical', the selected model has to include a
% spherical joint for the childBOdyName.
% 2) The scripts are not able to update the conditional via points. At the
% moment this is done manually afterwards. 
% ----------------------------------------------------------------------- % 
% Last modifications:
% 2018-04-23, Luca Modenese
% removed MD-Paedigree dependencies.
% ----------------------------------------------------------------------- % 

clear; clc

% add functions
addpath(genpath('./Matlab_functions_upd_coord_names'))

% select the model
[FileName,PathName] = uigetfile('*.osim','Please select the osim model of which you want to update joint and coordinate names.');

% importing libraries
import org.opensim.modeling.*

% read model in
osimModel = Model(fullfile(PathName,FileName));

% updating pelvis-ground joint
osimModel = updBuilderJoints2Osim(osimModel, 'pelvis', 'ground_pelvis',  'free', 'zxy', { 'pelvis_tilt','pelvis_list','pelvis_rotation',...
                                  'pelvis_tx','pelvis_ty','pelvis_tz'});

% loop through sides on the legs to update them. Assumes bilateral model
% but can easily be modified to handle mono-lateral.
side_set = {'r','l'};
for n_s = 1:2
    side = side_set{n_s};
    % going through the joints of each side of the two-leg model
    osimModel = updBuilderJoints2Osim(osimModel, ['femur_',side], ['hip_',side],  'spherical', 'zxy', {['hip_flexion_',side], ['hip_adduction_',side], ['hip_rotation_',side]});
    osimModel = updBuilderJoints2Osim(osimModel, ['tibia_',side], ['knee_',side],  'hinge', 'zxy', ['knee_angle_',side]);
    osimModel = updBuilderJoints2Osim(osimModel, ['talus_',side], ['ankle_',side], 'hinge', 'zxy', ['ankle_angle_',side]);
    osimModel = updBuilderJoints2Osim(osimModel, ['calcn_',side], ['subtalar_',side],  'hinge', 'zxy', ['subtalar_angle_',side]);
    osimModel = updBuilderJoints2Osim(osimModel, ['toes_',side], ['mtp_',side],  'hinge', 'zxy', ['mtp_angle_',side]);
end

% deblank mucles (just ensuring not surprises).
osimModel = deblankMuscleNames(osimModel);

% new model name
[~, name, est] = fileparts(FileName);
upd_name = [name,'_upd_names',est];

% printing new model
osimModel.print(fullfile(PathName,upd_name));
display(['The model with the updated joint and coordinates names has been written in folder: ', PathName])

% remove libraries from path
rmpath(genpath('./Matlab_functions_upd_coord_names'))
